from __future__ import annotations

import json
from pathlib import Path

import numpy as np
import pandas as pd

from bidsmreye.quality_control import compute_displacement
from bidsmreye.quality_control import compute_robust_outliers


def create_basic_json():

    output_location = Path().resolve()
    output_location = output_location.joinpath("tests", "data", "bidsmreye")

    sidecar_name = output_location.joinpath(
        "sub-01",
        "func",
        "sub-01_task-nback_space-MNI152NLin2009cAsym_desc-bidsmreye_eyetrack.json",
    )

    content = {"SamplingFrequency": 0.14285714285714285}

    json.dump(content, open(sidecar_name, "w"), indent=4)


def bidsmreye_eyetrack():

    output_location = Path().resolve()
    output_location = output_location.joinpath("tests", "data", "bidsmreye")

    return output_location.joinpath(
        "sub-01",
        "func",
        "sub-01_task-nback_space-MNI152NLin2009cAsym_desc-bidsmreye_eyetrack.tsv",
    )


def create_basic_data():
    return {
        "eye1_x_coordinate": np.random.randn(400),
        "eye1_y_coordinate": np.random.randn(400),
    }


def create_data_with_outliers():

    data = create_basic_data()

    data["eye_timestamp"] = np.arange(400)

    eye1_x_coordinate = data["eye1_x_coordinate"]
    eye1_y_coordinate = data["eye1_y_coordinate"]

    data["eye1_x_coordinate"][200] = (
        eye1_x_coordinate.mean() + eye1_x_coordinate.std() * 4
    )
    data["eye1_y_coordinate"][200] = (
        eye1_y_coordinate.mean() - eye1_y_coordinate.std() * 5
    )
    data["eye1_x_coordinate"][50] = eye1_x_coordinate.mean() - eye1_x_coordinate.std() * 5
    data["eye1_y_coordinate"][50] = eye1_y_coordinate.mean() + eye1_y_coordinate.std() * 4

    return data


def create_confounds_tsv():

    confounds_tsv = bidsmreye_eyetrack()

    df = pd.DataFrame(create_data_with_outliers())

    df["displacement"] = compute_displacement(
        df["eye1_x_coordinate"],
        df["eye1_y_coordinate"],
    )
    df["eye1_x_outliers"] = compute_robust_outliers(
        df["eye1_x_coordinate"], outlier_type="Carling"
    )
    df["eye1_y_outliers"] = compute_robust_outliers(
        df["eye1_y_coordinate"], outlier_type="Carling"
    )
    df["displacement_outliers"] = compute_robust_outliers(
        df["displacement"], outlier_type="Carling"
    )

    cols = df.columns.tolist()
    cols.insert(0, cols.pop(cols.index("eye_timestamp")))
    df = df[cols]

    df.to_csv(confounds_tsv, sep="\t", index=False)



from __future__ import annotations

import pandas as pd

from .utils import bidsmreye_eyetrack
from .utils import create_confounds_tsv
from bidsmreye.visualize import visualize_eye_gaze_data


def test_visualize_eye_gaze_data():

    confounds_tsv = bidsmreye_eyetrack()

    create_confounds_tsv()

    eye_gaze_data = pd.read_csv(confounds_tsv, sep="\t")

    fig = visualize_eye_gaze_data(eye_gaze_data)

    fig.show()



from __future__ import annotations

from typing import Any

import pandas as pd
import plotly.graph_objs as go
from plotly.subplots import make_subplots

OPACITY = 1
LINE_WIDTH = 3
FONT_SIZE = dict(size=14)
GRID_COLOR = "grey"
LINE_COLOR = "rgb(0, 150, 175)"
BG_COLOR = "rgb(255,255,255)"
HEAT_MAP_COLOR = "gnbu"


def value_range(X: pd.Series) -> list[float]:
    return [-X.max() * 1.2, X.max() * 1.2]


def time_range(time_stamps: pd.Series) -> list[float]:
    return [time_stamps.min() - 3, time_stamps.max() + 3]


def visualize_eye_gaze_data(
    eye_gaze_data: pd.DataFrame,
) -> Any:

    fig = go.FigureWidget(
        make_subplots(
            rows=3,
            cols=4,
            horizontal_spacing=0.05,
            vertical_spacing=0.1,
            shared_xaxes="columns",
            specs=[
                [{"colspan": 2}, None, {"rowspan": 2, "colspan": 2}, None],
                [{"colspan": 2}, None, None, None],
                [{"colspan": 2}, None, {"colspan": 2}, None],
            ],
        )
    )

    # Plot input signal together with split output signal (X & Y)
    plot_time_series(fig, eye_gaze_data, title_text="X", row=1, col=1)
    plot_time_series(fig, eye_gaze_data, title_text="Y", row=2, col=1)
    plot_time_series(
        fig,
        eye_gaze_data,
        title_text="displacement",
        row=3,
        col=1,
        plotting_range=[-0.1, eye_gaze_data["displacement"].max() * 1.1],
        line_color="grey",
    )
    fig.update_xaxes(
        row=3,
        col=1,
        title=dict(text="Time (s)", standoff=16, font=FONT_SIZE),
        tickfont=FONT_SIZE,
    )

    plot_heat_map(fig, eye_gaze_data)

    return fig


def plot_time_series(
    fig: Any,
    eye_gaze_data: pd.DataFrame,
    title_text: str,
    row: int,
    col: int,
    plotting_range: list[float] | None = None,
    line_color: str = LINE_COLOR,
) -> None:

    outliers = None

    values_to_plot = eye_gaze_data["eye1_x_coordinate"]
    outliers = eye_gaze_data["eye1_x_outliers"]
    outlier_color = "orange"
    if title_text == "Y":
        values_to_plot = eye_gaze_data["eye1_y_coordinate"]
        outliers = eye_gaze_data["eye1_y_outliers"]
    elif title_text == "displacement":
        values_to_plot = eye_gaze_data["displacement"]
        outliers = eye_gaze_data["displacement_outliers"]
        outlier_color = "red"

    if plotting_range is None:
        plotting_range = value_range(values_to_plot)

    fig.add_trace(
        go.Scatter(
            x=time_range(eye_gaze_data["eye_timestamp"]),
            y=[0, 0],
            mode="lines",
            line_color="black",
            opacity=OPACITY,
            line_width=LINE_WIDTH - 1,
        ),
        row=row,
        col=col,
    )

    fig.add_trace(
        go.Scatter(
            x=eye_gaze_data["eye_timestamp"],
            y=values_to_plot,
            mode="lines",
            line_color=line_color,
            opacity=OPACITY,
            line_width=LINE_WIDTH,
        ),
        row=row,
        col=col,
    )

    if outliers is not None:
        fig.add_trace(
            go.Scatter(
                x=eye_gaze_data["eye_timestamp"][outliers == 1],
                y=values_to_plot[outliers == 1],
                mode="markers",
                marker_color=outlier_color,
                marker_size=10,
                opacity=OPACITY,
            ),
            row=row,
            col=col,
        )

    fig.update_xaxes(
        range=time_range(eye_gaze_data["eye_timestamp"]),
        row=row,
        col=col,
        gridcolor=GRID_COLOR,
        tickfont=FONT_SIZE,
    )
    fig.update_yaxes(
        range=plotting_range,
        row=row,
        col=col,
        gridcolor=GRID_COLOR,
        ticksuffix="°",
        title=dict(text=title_text, standoff=0, font=FONT_SIZE),
        tickfont=FONT_SIZE,
    )

    fig.update_layout(showlegend=False, plot_bgcolor=BG_COLOR, paper_bgcolor=BG_COLOR)


def plot_heat_map(fig: Any, eye_gaze_data: pd.DataFrame) -> None:

    X = eye_gaze_data["eye1_x_coordinate"]
    Y = eye_gaze_data["eye1_y_coordinate"]

    x_range = value_range(X)
    y_range = value_range(Y)

    fig.add_trace(
        go.Histogram2dContour(x=X, y=Y, colorscale=HEAT_MAP_COLOR),
        row=1,
        col=3,
    )

    fig.add_trace(
        go.Scatter(
            x=x_range,
            y=[0, 0],
            mode="lines",
            line_color="black",
            opacity=OPACITY,
            line_width=LINE_WIDTH - 2,
        ),
        row=1,
        col=3,
    )
    fig.add_trace(
        go.Scatter(
            x=[0, 0],
            y=y_range,
            mode="lines",
            line_color="black",
            opacity=OPACITY,
            line_width=LINE_WIDTH - 2,
        ),
        row=1,
        col=3,
    )

    fig.add_trace(
        go.Scatter(
            x=X,
            y=Y,
            opacity=0.4,
            line=dict(color="black", width=1, dash="dash"),
        ),
        row=1,
        col=3,
    )

    outliers = eye_gaze_data["eye1_x_outliers"]
    outlier_color = "orange"
    add_outliers_to_heatmap(fig, X, Y, outliers, outlier_color)
    outliers = eye_gaze_data["eye1_y_outliers"]
    add_outliers_to_heatmap(fig, X, Y, outliers, outlier_color)
    outliers = eye_gaze_data["displacement_outliers"]
    outlier_color = "red"
    add_outliers_to_heatmap(fig, X, Y, outliers, outlier_color)

    fig.update_xaxes(
        row=1,
        col=3,
        range=value_range(X),
        ticksuffix="°",
        title=dict(text="X", standoff=16, font=FONT_SIZE),
        tickfont=FONT_SIZE,
    )
    fig.update_yaxes(
        row=1,
        col=3,
        range=value_range(Y),
        ticksuffix="°",
        title=dict(text="Y", standoff=16, font=FONT_SIZE),
        tickfont=FONT_SIZE,
    )

    fig.update_layout(showlegend=False)


def add_outliers_to_heatmap(
    fig: Any, X: pd.Series, Y: pd.Series, outliers: pd.Series, outlier_color: str
) -> None:
    fig.add_trace(
        go.Scatter(
            x=X[outliers == 1],
            y=Y[outliers == 1],
            mode="markers",
            marker_color=outlier_color,
            marker_size=8,
            opacity=OPACITY,
        ),
        row=1,
        col=3,
    )



from __future__ import annotations

from pathlib import Path

from bidsmreye.prepare_data import combine_data_with_empty_labels
from bidsmreye.utils import get_dataset_layout


def test_combine_data_with_empty_labels():

    output_location = Path().resolve()
    output_location = output_location.joinpath("tests", "data", "bidsmreye")

    layout_out = get_dataset_layout(output_location)

    file = output_location.joinpath(
        "sub-01",
        "func",
        "sub-01_task-nback_run-01_space-MNI152NLin2009cAsym_desc-eye_mask.p",
    )

    no_label_file = combine_data_with_empty_labels(layout_out, file)

    assert no_label_file.is_file()

    expected_file = output_location.joinpath(
        "sub-01",
        "func",
        "sub-01_task-nback_run-01_space-MNI152NLin2009cAsym_desc-nolabel_bidsmreye.npz",
    )

    assert no_label_file == expected_file

    no_label_file.unlink()



from __future__ import annotations

import shutil
from pathlib import Path

import pytest
from bids.tests import get_test_data_path

from bidsmreye.utils import Config
from bidsmreye.utils import config_to_dict
from bidsmreye.utils import copy_license
from bidsmreye.utils import create_bidsname
from bidsmreye.utils import get_bidsname_config
from bidsmreye.utils import get_config
from bidsmreye.utils import get_dataset_layout
from bidsmreye.utils import get_deepmreye_filename
from bidsmreye.utils import get_pybids_config
from bidsmreye.utils import init_derivatives_layout
from bidsmreye.utils import list_subjects
from bidsmreye.utils import return_deepmreye_output_filename
from bidsmreye.utils import return_regex


def pybids_test_dataset():
    return Path(get_test_data_path()).joinpath("synthetic", "derivatives", "fmriprep")


def test_Config():

    cfg = Config(
        pybids_test_dataset(),
        Path(__file__).parent.joinpath("data"),
    )
    assert not cfg.debug
    assert not cfg.non_linear_coreg
    assert cfg.input_folder == pybids_test_dataset()
    assert cfg.output_folder == Path(__file__).parent.joinpath("data", "bidsmreye")
    assert sorted(cfg.participant) == ["01", "02", "03", "04", "05"]
    assert sorted(cfg.task) == ["nback", "rest"]
    assert sorted(cfg.space) == ["MNI152NLin2009cAsym", "T1w"]


def test_config_to_dict_smoke():
    cfg = Config(
        pybids_test_dataset(),
        Path(__file__).parent.joinpath("data"),
    )
    cfg_dict = config_to_dict(cfg)


def test_Config_task_omit_missing_values():
    cfg = Config(
        pybids_test_dataset(),
        Path(__file__).parent.joinpath("data"),
        task=["auditory", "rest"],
        participant=["01", "07"],
        space=["T1w", "T2w"],
    )
    assert cfg.participant == ["01"]
    assert cfg.task == ["rest"]
    assert cfg.space == ["T1w"]


def test_missing_subject():
    with pytest.warns(UserWarning):
        Config(
            pybids_test_dataset(),
            Path(__file__).parent.joinpath("data"),
            participant=["01", "07"],
        )


def test_no_subject():
    with pytest.raises(Exception) as e_info:
        Config(
            pybids_test_dataset(),
            Path(__file__).parent.joinpath("data"),
            participant=["99"],
        )
    assert e_info.type == RuntimeError


def test_missing_task():
    with pytest.warns(UserWarning):
        Config(
            pybids_test_dataset(),
            Path(__file__).parent.joinpath("data"),
            task=["auditory", "rest"],
        )


def test_no_task():
    with pytest.raises(Exception) as e_info:
        Config(
            pybids_test_dataset(),
            Path(__file__).parent.joinpath("data"),
            task=["foo"],
        )
    assert e_info.type == RuntimeError


def test_missing_space():
    with pytest.warns(UserWarning):
        Config(
            pybids_test_dataset(),
            Path(__file__).parent.joinpath("data"),
            space=["T1w", "T2w"],
        )


def test_no_subject():
    with pytest.raises(Exception) as e_info:
        Config(
            pybids_test_dataset(),
            Path(__file__).parent.joinpath("data"),
            space=["T2w"],
        )
    assert e_info.type == RuntimeError


def test_return_regex():
    assert return_regex("foo") == "^foo$"
    assert return_regex("^foo") == "^foo$"
    assert return_regex("foo$") == "^foo$"
    assert return_regex(["foo", "bar"]) == "^foo$|^bar$"


def test_list_subjects():

    cfg = Config(
        pybids_test_dataset(),
        Path(__file__).parent.joinpath("data"),
    )

    layout = get_dataset_layout(pybids_test_dataset())

    subjects = list_subjects(cfg, layout)
    assert len(subjects) == 5


def test_get_dataset_layout_smoke_test():
    get_dataset_layout(Path("data"))

    shutil.rmtree("data")


def test_get_deepmreye_filename():

    layout = get_dataset_layout(pybids_test_dataset())

    output_file = Path(get_test_data_path()).joinpath(
        "synthetic",
        "derivatives",
        "fmriprep",
        "sub-01",
        "ses-01",
        "func",
        "mask_sub-01_ses-01_task-nback_run-01_space-MNI152NLin2009cAsym_desc-preproc_bold.p",
    )

    img = layout.get(
        return_type="filename",
        subject="01",
        suffix="bold",
        task="nback",
        space="MNI152NLin2009cAsym",
        extension=".nii.gz",
    )
    deepmreye_mask_name = get_deepmreye_filename(layout, img, "mask")

    assert deepmreye_mask_name == output_file


def test_return_deepmreye_output_filename():

    input_file = "sub-01_task-rest_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
    output_filename = return_deepmreye_output_filename(input_file, "mask")
    expected_output_file = (
        "mask_sub-01_task-rest_space-MNI152NLin2009cAsym_desc-preproc_bold.p"
    )
    assert output_filename == expected_output_file

    input_file = "sub-01_task-rest_space-MNI152NLin2009cAsym_desc-preproc_bold.nii"
    output_filename = return_deepmreye_output_filename(input_file, "mask")
    assert output_filename == expected_output_file


def test_get_config_error():
    with pytest.raises(FileNotFoundError):
        get_config("", "foo.json")


def test_get_bidsname_config_smoke():
    cfg = get_bidsname_config()
    assert cfg is not None


def test_get_pybids_config_smoke():
    cfg = get_pybids_config()
    assert cfg is not None


def test_write_dataset_description():

    output_location = Path().resolve()
    output_location = Path.joinpath(output_location, "derivatives")

    init_derivatives_layout(output_location)

    shutil.rmtree(output_location)


def test_copy_license():

    output_dir = Path().resolve()
    output_location = output_dir.joinpath("derivatives")

    shutil.rmtree(output_location, ignore_errors=True)

    license_file = copy_license(output_location)

    assert license_file.is_file()
    assert str(license_file) == str(output_location.joinpath("LICENSE"))

    copy_license(output_location)

    shutil.rmtree(output_location)


def test_create_bidsname():

    output_location = Path().resolve()
    output_location = Path.joinpath(output_location, "derivatives")

    layout = get_dataset_layout(output_location)
    filename = Path("inputs").joinpath(
        "raw",
        "sub-01",
        "ses-01",
        "func",
        "sub-01_ses-01_task-motion_run-1_bold.nii",
    )

    output_file = create_bidsname(layout, filename=filename, filetype="mask")

    rel_path = output_file.relative_to(layout.root)

    assert rel_path == Path("sub-01").joinpath(
        "ses-01", "func", "sub-01_ses-01_task-motion_run-1_desc-eye_mask.p"
    )

    shutil.rmtree(output_location)



from __future__ import annotations

import shutil
from pathlib import Path

from bidsmreye.methods import methods


def test_methods():

    output_dir = Path.cwd().joinpath("temp")
    output_file = methods(output_dir=output_dir)

    assert output_dir.is_dir()
    assert output_file.is_file()

    shutil.rmtree(output_dir)


def test_methods_calibration_data():

    output_dir = Path.cwd().joinpath("temp")
    output_file = methods(output_dir=output_dir, model_name="calibration_data")

    assert output_dir.is_dir()
    assert output_file.is_file()

    shutil.rmtree(output_dir)



from __future__ import annotations

import shutil
from pathlib import Path

import pytest

from bidsmreye.download import download
from bidsmreye.download import download_parser


def test_download_parser():

    parser = download_parser()

    assert parser.description == "Download deepmreye pretrained model from OSF."

    args, unknowns = parser.parse_known_args(
        [
            "--model_name",
            "1_guided_fixations",
            "--output_dir",
            "/home/bob/models",
        ]
    )

    assert args.output_dir == "/home/bob/models"


def test_download():

    output_dir = Path().joinpath("tmp")

    download(model_name="1_guided_fixations", output_dir=str(output_dir))

    assert output_dir.is_dir()
    assert output_dir.joinpath("dataset_1_guided_fixations.h5").is_file()

    shutil.rmtree(output_dir)


def test_download_basic():

    download()
    output_file = download()

    output_dir = Path.cwd().joinpath("models")
    print(output_file)
    assert output_dir.is_dir()
    assert output_dir.joinpath("dataset_1to6.h5").is_file()

    shutil.rmtree(output_dir)


def test_download_unknown_model():

    with pytest.warns(UserWarning):
        download(model_name="foo")



"""Download the models from OSF."""
from __future__ import annotations

import argparse
import sys
import warnings
from pathlib import Path
from typing import Any
from typing import IO

import pkg_resources
import pooch
import rich

from bidsmreye.utils import available_models
from bidsmreye.utils import bidsmreye_log
from bidsmreye.utils import default_model

log = bidsmreye_log(name="bidsmreye")


class MuhParser(argparse.ArgumentParser):
    """Parser for the main script."""

    def _print_message(self, message: str, file: IO[str] | None = None) -> None:
        rich.print(message, file=file)


def download_parser() -> MuhParser:
    """Execute the main script."""
    parser = MuhParser(
        description="Download deepmreye pretrained model from OSF.",
        epilog="""
        For a more readable version of this help section,
        see the online https://bidsmreye.readthedocs.io/.
        """,
    )
    parser.add_argument(
        "--model_name",
        help="""
        Model to download.
        """,
        choices=available_models(),
        default=default_model(),
    )
    parser.add_argument(
        "--output_dir",
        help="""
        The directory where the model files will be stored.
        """,
        default=Path.cwd().joinpath("models"),
    )

    return parser


def cli(argv: Any = sys.argv) -> None:
    """Download the models from OSF.

    :return: _description_
    :rtype: _type_
    """
    parser = download_parser()
    args = parser.parse_args(argv[1:])

    download(model_name=args.model_name, output_dir=args.output_dir)


def download(
    model_name: str | Path | None = None, output_dir: Path | str | None = None
) -> Path | None:
    """Download the models from OSF.

    :param model_name: Model to download. defaults to None
    :type model_name: str, optional

    :param output_dir: Path where to save the model. Defaults to None.
    :type output_dir: Path, optional

    :return: Path to the downloaded model.
    :rtype: Path
    """
    if not model_name:
        model_name = default_model()
    if isinstance(model_name, Path):
        assert model_name.is_file()
        return model_name.resolve()
    if model_name not in available_models():
        warnings.warn(f"{model_name} is not a valid model name.")
        return None

    if output_dir is None:
        output_dir = Path.cwd().joinpath("models")
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)

    POOCH = pooch.create(
        path=output_dir,
        base_url="https://osf.io/download/",
        registry=None,
    )
    registry_file = pkg_resources.resource_stream("bidsmreye", "models/registry.txt")
    POOCH.load_registry(registry_file)

    output_file = output_dir.joinpath(f"dataset_{model_name}")

    if not output_file.is_file():

        file_idx = available_models().index(model_name)
        filename = f"dataset_{available_models()[file_idx]}.h5"
        output_file = POOCH.fetch(filename, progressbar=True)
        if isinstance(output_file, str):
            output_file = Path(output_file)

    else:
        log.info(f"{output_file} already exists.")

    return output_file



from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd

from .utils import bidsmreye_eyetrack
from .utils import create_basic_data
from .utils import create_basic_json
from .utils import create_confounds_tsv
from bidsmreye.quality_control import add_qc_to_sidecar
from bidsmreye.quality_control import compute_robust_outliers
from bidsmreye.quality_control import perform_quality_control
from bidsmreye.quality_control import quality_control
from bidsmreye.utils import Config
from bidsmreye.utils import get_dataset_layout


def time_series():

    return [
        0.6876,
        0.9751,
        0.1322,
        0.2420,
        1.4233,
        1.2617,
        -0.8619,
        -0.9471,
        2.6217,
        -0.6192,
        -1.0646,
        -0.4872,
        -0.1146,
        0.3007,
        -0.4089,
        0.1137,
        -0.0946,
        0.7829,
        1.8999,
        -1.0088,
    ]


def test_compute_robust_outliers():

    outliers = compute_robust_outliers(pd.Series(time_series()))

    expected_outliers = [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    assert outliers == expected_outliers


def test_compute_robust_outliers_carling():

    series = time_series()
    series[8] = 10
    outliers = compute_robust_outliers(pd.Series(series), outlier_type="Carling")

    expected_outliers = [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    assert outliers == expected_outliers


def test_compute_robust_outliers_carling_with_nan():

    series = time_series()
    series[1] = np.nan
    series[8] = 10
    outliers = compute_robust_outliers(pd.Series(series), outlier_type="Carling")

    expected_outliers = [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    assert outliers == expected_outliers


def test_compute_robust_with_nan():

    series = time_series()
    series[1] = np.nan
    series[8] = 10

    outliers = compute_robust_outliers(pd.Series(series))

    expected_outliers = [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    assert outliers == expected_outliers


def test_quality_control():

    create_basic_json()

    output_location = Path().resolve()
    output_location = output_location.joinpath("tests", "data")

    confounds_tsv = bidsmreye_eyetrack()

    df = pd.DataFrame(create_basic_data())
    df.to_csv(confounds_tsv, sep="\t", index=False)

    cfg = Config(
        output_location,
        output_location,
    )

    quality_control(cfg)


def test_perform_quality_control():

    create_basic_json()

    output_location = Path().resolve()
    output_location = output_location.joinpath("tests", "data", "bidsmreye")
    layout_out = get_dataset_layout(output_location)
    confounds_tsv = bidsmreye_eyetrack()

    df = pd.DataFrame(create_basic_data())
    df.to_csv(confounds_tsv, sep="\t", index=False)

    perform_quality_control(layout_out, confounds_tsv)


def test_add_qc_to_sidecar():

    create_basic_json()

    create_confounds_tsv()

    output_location = Path().resolve()
    output_location = output_location.joinpath("tests", "data", "bidsmreye")
    layout_out = get_dataset_layout(output_location)
    confounds_tsv = bidsmreye_eyetrack()

    add_qc_to_sidecar(layout_out, confounds_tsv)



"""Run coregistration and extract data."""
from __future__ import annotations

import logging
import pickle
from pathlib import Path
from typing import Any

import nibabel as nib
import numpy as np
from bids import BIDSLayout  # type: ignore
from deepmreye import preprocess

from bidsmreye.methods import methods
from bidsmreye.utils import check_layout
from bidsmreye.utils import Config
from bidsmreye.utils import config_to_dict
from bidsmreye.utils import copy_license
from bidsmreye.utils import create_bidsname
from bidsmreye.utils import create_dir_if_absent
from bidsmreye.utils import create_sidecar
from bidsmreye.utils import get_dataset_layout
from bidsmreye.utils import get_deepmreye_filename
from bidsmreye.utils import list_subjects
from bidsmreye.utils import move_file
from bidsmreye.utils import set_dataset_description
from bidsmreye.utils import set_this_filter
from bidsmreye.utils import write_dataset_description


log = logging.getLogger("bidsmreye")


def coregister_and_extract_data(img: str, non_linear_coreg: bool = False) -> None:
    """Coregister image to eye template and extract data from eye mask for one image.

    :param img: Image to coregister and extract data from
    :type img: str
    """
    (
        eyemask_small,
        eyemask_big,
        dme_template,
        mask,
        x_edges,
        y_edges,
        z_edges,
    ) = preprocess.get_masks()

    transforms = ["Affine", "Affine", "SyNAggro"] if non_linear_coreg else None

    preprocess.run_participant(
        img,
        dme_template,
        eyemask_big,
        eyemask_small,
        x_edges,
        y_edges,
        z_edges,
        transforms=transforms,
    )


def combine_data_with_empty_labels(layout_out: BIDSLayout, img: Path, i: int = 1) -> Path:
    """Combine data with empty labels.

    :param layout_out: _description_
    :type layout_out: _type_

    :param subject_label: _description_
    :type subject_label: _type_

    :param img: _description_
    :type img: _type_

    :param i: _description_, defaults to 1
    :type i: int, optional
    """
    log.debug(f"Combining data with empty labels: {img}")

    # Load data and normalize it
    data = pickle.load(open(img, "rb"))
    data = preprocess.normalize_img(data)

    # If experiment has no labels use dummy labels
    # 10 is the number of subTRs used in the pretrained weights, 2 is XY
    labels = np.zeros((data.shape[3], 10, 2))

    entities = layout_out.parse_file_entities(img)

    # Store for each runs
    subj: dict[str, list[Any]] = {"data": [], "labels": [], "ids": []}
    subj["data"].append(data)
    subj["labels"].append(labels)
    subj["ids"].append(([entities["subject"]] * labels.shape[0], [i] * labels.shape[0]))

    output_file = create_bidsname(layout_out, Path(img), "no_label")

    preprocess.save_data(
        output_file.name,
        subj["data"],
        subj["labels"],
        subj["ids"],
        layout_out.root,
        center_labels=False,
    )

    file_to_move = Path(layout_out.root).joinpath("..", "bidsmreye", output_file.name)

    move_file(file_to_move, output_file)

    return output_file


def process_subject(
    cfg: Config, layout_in: BIDSLayout, layout_out: BIDSLayout, subject_label: str
) -> None:
    """Run coregistration and extract data for one subject.

    :param cfg: Configuration object.
    :type cfg: Config

    :param layout_in: Layout input dataset.
    :type layout_in: BIDSLayout

    :param layout_out: Layout output dataset.
    :type layout_out: BIDSLayout

    :param subject_label: Can be a regular expression.
    :type subject_label: str
    """
    log.info(f"Running subject: {subject_label}")

    this_filter = set_this_filter(cfg, subject_label, "bold")

    bf = layout_in.get(
        return_type="filename",
        regex_search=True,
        **this_filter,
    )

    to_print = [str(Path(x).relative_to(layout_in.root)) for x in bf]
    log.debug(f"Found files\n{to_print}")

    for img in bf:

        log.info(f"Processing file: {Path(img).name}")

        coregister_and_extract_data(img, non_linear_coreg=cfg.non_linear_coreg)

        report_name = create_bidsname(layout_out, img, "report")
        deepmreye_mask_report = get_deepmreye_filename(layout_in, img, "report")
        move_file(deepmreye_mask_report, report_name)

        mask_name = create_bidsname(layout_out, img, "mask")
        deepmreye_mask_name = get_deepmreye_filename(layout_in, img, "mask")
        move_file(deepmreye_mask_name, mask_name)

        save_sampling_frequency_to_json(layout_out, img)

        combine_data_with_empty_labels(layout_out, mask_name)


def save_sampling_frequency_to_json(layout_out: BIDSLayout, img: str) -> None:
    func_img = nib.load(img)
    header = func_img.header
    repetition_time = header.get_zooms()[3]
    if repetition_time <= 1:
        log.warning(f"Found a repetition time of {repetition_time} seconds.")
    create_sidecar(layout_out, img, SamplingFrequency=1 / float(repetition_time))


def prepare_data(cfg: Config) -> None:
    """Run coregistration and extract data for all subjects.

    :param cfg: Configuration object
    :type cfg: Config
    """
    log.info("PREPARING DATA")

    layout_in = get_dataset_layout(cfg.input_folder, use_database=True)
    check_layout(cfg, layout_in)

    create_dir_if_absent(cfg.output_folder)
    layout_out = get_dataset_layout(cfg.output_folder)
    # TODO add readme and license
    layout_out = set_dataset_description(layout_out)
    layout_out.dataset_description["DatasetType"] = "derivative"
    layout_out.dataset_description["GeneratedBy"][0]["Name"] = "bidsmreye"
    layout_out.dataset_description["GeneratedBy"][0]["config"] = config_to_dict(cfg)
    write_dataset_description(layout_out)

    citation_file = methods(cfg.output_folder)
    log.info(f"Method section generated: {citation_file}")

    license_file = copy_license(cfg.output_folder)
    log.info(f"License file added: {license_file}")

    subjects = list_subjects(cfg, layout_in)

    if cfg.non_linear_coreg:
        log.debug("Using non-linear coregistration")

    for subject_label in subjects:

        process_subject(cfg, layout_in, layout_out, subject_label)



"""TODO."""
from __future__ import annotations

import json
import logging
import math
from pathlib import Path

import numpy as np
import pandas as pd
from bids import BIDSLayout  # type: ignore
from scipy.stats.distributions import chi2

from bidsmreye.utils import check_layout
from bidsmreye.utils import Config
from bidsmreye.utils import create_bidsname
from bidsmreye.utils import create_dir_for_file
from bidsmreye.utils import get_dataset_layout
from bidsmreye.utils import list_subjects
from bidsmreye.utils import set_this_filter
from bidsmreye.visualize import visualize_eye_gaze_data

log = logging.getLogger("bidsmreye")


def compute_displacement(x: pd.Series, y: pd.Series) -> pd.Series:

    return np.sqrt((x.diff() ** 2) + (y.diff() ** 2))


def add_qc_to_sidecar(layout: BIDSLayout, confounds_tsv: str | Path) -> Path:
    """_summary_

    :param layout: _description_
    :type layout: BIDSLayout

    :param confounds_tsv: _description_
    :type confounds_tsv: str | Path

    :return: _description_
    :rtype: Path
    """
    confounds_tsv = Path(confounds_tsv)
    confounds = pd.read_csv(confounds_tsv, sep="\t")

    sidecar_name = create_bidsname(layout, confounds_tsv, "confounds_json")

    with open(sidecar_name) as f:
        content = json.load(f)

        content["NbDisplacementOutliers"] = confounds["displacement_outliers"].sum()
        content["NbXOutliers"] = confounds["eye1_x_outliers"].sum()
        content["NbYOutliers"] = confounds["eye1_y_outliers"].sum()
        content["eye1XVar"] = confounds["eye1_x_coordinate"].var()
        content["eye1YVar"] = confounds["eye1_y_coordinate"].var()

    json.dump(content, open(sidecar_name, "w"), indent=4)

    return sidecar_name


def perform_quality_control(layout: BIDSLayout, confounds_tsv: str | Path) -> None:
    """Perform quality control on the confounds.

    :param layout: pybids layout to of the dataset to act on.
    :type layout: BIDSLayout

    :param confounds_tsv: Path to the confounds TSV file.
    :type confounds_tsv: str | Path
    """
    confounds_tsv = Path(confounds_tsv)
    confounds = pd.read_csv(confounds_tsv, sep="\t")

    repetition_time = get_repetition_time(layout, confounds_tsv)
    nb_timepoints = confounds.shape[0]
    eye_timestamp = np.arange(0, repetition_time * nb_timepoints, repetition_time)
    confounds["eye_timestamp"] = eye_timestamp

    cols = confounds.columns.tolist()
    cols.insert(0, cols.pop(cols.index("eye_timestamp")))
    confounds = confounds[cols]

    confounds["displacement"] = compute_displacement(
        confounds["eye1_x_coordinate"], confounds["eye1_y_coordinate"]
    )

    confounds["displacement_outliers"] = compute_robust_outliers(
        confounds["displacement"], outlier_type="Carling"
    )
    log.debug(f"Found {confounds['displacement_outliers'].sum()} displacement outliers")

    confounds["eye1_x_outliers"] = compute_robust_outliers(
        confounds["eye1_x_coordinate"], outlier_type="Carling"
    )
    log.debug(f"Found {confounds['eye1_x_outliers'].sum()} x outliers")

    confounds["eye1_y_outliers"] = compute_robust_outliers(
        confounds["eye1_y_coordinate"], outlier_type="Carling"
    )
    log.debug(f"Found {confounds['eye1_y_outliers'].sum()} y outliers")

    confounds.to_csv(confounds_tsv, sep="\t", index=False)

    sidecar_name = add_qc_to_sidecar(layout, confounds_tsv)
    log.info(f"Quality control data added to {sidecar_name}")

    fig = visualize_eye_gaze_data(confounds)
    if log.isEnabledFor(logging.DEBUG):
        fig.show()
    visualization_html_file = create_bidsname(layout, confounds_tsv, "confounds_html")
    create_dir_for_file(visualization_html_file)
    fig.write_html(visualization_html_file)


def get_repetition_time(layout: BIDSLayout, file: str | Path) -> float | None:
    """Get the repetition time from the sidecar JSON file."""
    repetition_time = None

    sidecar_name = create_bidsname(layout, file, "confounds_json")

    if sidecar_name.is_file():
        with open(sidecar_name) as f:
            content = json.load(f)
            SamplingFrequency = content.get("SamplingFrequency", None)
            if SamplingFrequency is not None and SamplingFrequency > 0:
                repetition_time = 1 / SamplingFrequency

    return repetition_time


def quality_control(cfg: Config) -> None:
    """Run quality control on the output dataset."""

    log.info("QUALITY CONTROL")

    layout_out = get_dataset_layout(cfg.output_folder)
    check_layout(cfg, layout_out)

    subjects = list_subjects(cfg, layout_out)

    for subject_label in subjects:

        qc_subject(cfg, layout_out, subject_label)


def qc_subject(cfg: Config, layout_out: BIDSLayout, subject_label: str) -> None:
    """Run quality control for one subject."""

    log.info(f"Running subject: {subject_label}")

    this_filter = set_this_filter(cfg, subject_label, "eyetrack")

    data = layout_out.get(
        return_type="filename",
        regex_search=True,
        **this_filter,
    )

    to_print = [str(Path(x).relative_to(layout_out.root)) for x in data]
    log.debug(f"Found files\n{to_print}")

    for file in data:

        perform_quality_control(layout_out, file)


def compute_robust_outliers(
    time_series: pd.Series, outlier_type: str | None = None
) -> list[int] | type[NotImplementedError]:
    """Compute robust ouliers of a time series using S-outliers or Carling's k.

    :param series: Series to compute the outliers on.
    :type series: pd.Series

    :param outlier_type: Default to 'S-outliers'.
    :type outlier_type: str

    :return: Series of booleans indicating the outliers.
    :rtype: pd.Series

    Adapted from spmup:
    https://github.com/CPernet/spmup/blob/master/QA/spmup_comp_robust_outliers.m

    S-outliers is the default options, it is independent of a measure of
    centrality as this is based on the median of pair-wise distances. This is
    a very sensitive measures, i.e. it has a relatively high false positive
    rates. As such it is a great detection tools.

    The adjusted Carling's box-plot rule can also be used, and derived from
    the median of the data: outliers are outside the bound of median +/- k*IQR,
    with k = (17.63*n-23.64)/(7.74*n-3.71). This is a more specific measure,
    as such it is 'better' than S-outliers to regress-out, removing bad data
    points (assuming we don't want to 'remove' too many).

    References:

    - Rousseeuw, P. J., and Croux, C. (1993). Alternatives to the the median
    absolute deviation. J. Am. Stat. Assoc. 88, 1273-1263.
    <https://www.tandfonline.com/doi/abs/10.1080/01621459.1993.10476408>

    - Carling, K. (2000). Resistant outlier rules and the non-Gaussian case.
    Stat. Data Anal. 33, 249:258.
    <http://www.sciencedirect.com/science/article/pii/S0167947399000572>

    - Hoaglin, D.C., Iglewicz, B. (1987) Fine-tuning some resistant rules for
    outlier labelling. J. Amer. Statist. Assoc., 82 , 1147:1149
    <http://www.tandfonline.com/doi/abs/10.1080/01621459.1986.10478363>
    """

    if outlier_type is None:
        outlier_type = "S-outliers"

    if outlier_type == "S-outliers":

        k = np.sqrt(chi2.ppf(0.975, df=1))

        non_nan_idx = time_series.index[~time_series.isnull()].tolist()

        distance = []
        for i in non_nan_idx:

            this_timepoint = time_series[i]

            # all but current data point
            indices = list(range(len(time_series)))
            indices.pop(i)

            tmp = time_series[indices]
            tmp.dropna(inplace=True)

            # median of all pair-wise distances
            distance.append(np.median(abs(this_timepoint - tmp)))

        # get the S estimator
        consistency_factor = 1.1926
        Sn = consistency_factor * np.median(distance)

        # get the outliers in a normal distribution
        # no scaling needed as S estimates already std(data)

        outliers = np.zeros(len(time_series))
        outliers[non_nan_idx] = (distance / Sn) > k

        return outliers.tolist()

    elif outlier_type == "Carling":

        # interquartile range
        nan_less = time_series.dropna()
        nb_timepoints = len(nan_less)
        y = sorted(nan_less)
        j = math.floor(nb_timepoints / 4 + 5 / 12)
        g = (nb_timepoints / 4) - j + (5 / 12)
        k = nb_timepoints - j + 1

        lower_quartiles = (1 - g) * y[j] + g * y[j + 1]
        higher_quartiles = (1 - g) * y[k] + g * y[k - 1]
        inter_quartiles_range = higher_quartiles - lower_quartiles

        # robust outliers
        M = np.median(nan_less)
        carling_k = (17.63 * nb_timepoints - 23.64) / (7.74 * nb_timepoints - 3.71)
        lt = time_series < (M - carling_k * inter_quartiles_range)
        gt = time_series > (M + carling_k * inter_quartiles_range)
        df = pd.DataFrame({"lt": lt, "gt": gt})
        outliers = df["lt"] | df["gt"]

        return list(map(float, outliers))  # type: ignore

    else:
        raise ValueError(f"Unknown outlier_type: {outlier_type}")



#!/usr/bin/env python
"""Main script."""
from __future__ import annotations

import argparse
import sys
from typing import Any
from typing import IO

import rich

from . import _version
from bidsmreye.download import download
from bidsmreye.generalize import generalize
from bidsmreye.prepare_data import prepare_data
from bidsmreye.utils import available_models
from bidsmreye.utils import bidsmreye_log
from bidsmreye.utils import Config
from bidsmreye.utils import default_log_level
from bidsmreye.utils import default_model
from bidsmreye.utils import log_levels

__version__ = _version.get_versions()["version"]

log = bidsmreye_log(name="bidsmreye")


def cli(argv: Any = sys.argv) -> None:
    """Run the bids app.

    :param argv: _description_, defaults to sys.argv
    :type argv: _type_, optional
    """
    parser = common_parser()

    args = parser.parse_args(argv[1:])

    model_weights_file = args.model or None

    cfg = Config(
        args.bids_dir,
        args.output_dir,
        participant=args.participant_label or None,
        space=args.space or None,
        task=args.task or None,
        run=args.run or None,
        debug=args.debug,
        model_weights_file=model_weights_file,
        reset_database=args.reset_database,
        bids_filter=args.bids_filter_file,
        non_linear_coreg=args.non_linear_coreg,
    )  # type: ignore

    # TODO integrate as part of base config
    # https://stackoverflow.com/a/53293042/14223310
    log_level = log_levels().index(default_log_level())
    # For each "-v" flag, adjust the logging verbosity accordingly
    # making sure to clamp off the value from 0 to 4, inclusive of both
    for adjustment in args.log_level or ():
        log_level = min(len(log_levels()) - 1, max(log_level + adjustment, 0))
    log_level_name = log_levels()[log_level]
    log.setLevel(log_level_name)

    log.info("Running bidsmreye version %s", __version__)

    if cfg.debug:
        log.debug("DEBUG MODE")

    log.debug(f"args:\n{args}")
    log.debug(f"Configuration:\n{cfg}")

    if args.action in ["all", "generalize"] and isinstance(cfg.model_weights_file, str):
        cfg.model_weights_file = download(cfg.model_weights_file)

    if args.analysis_level == "participant":

        if args.action == "all":
            prepare_data(cfg)
            generalize(cfg)

        elif args.action == "prepare":
            prepare_data(cfg)

        elif args.action == "generalize":
            generalize(cfg)

        else:
            log.error("Unknown action")
            sys.exit(1)


class MuhParser(argparse.ArgumentParser):
    """Parser for the main script."""

    def _print_message(self, message: str, file: IO[str] | None = None) -> None:
        rich.print(message, file=file)


def common_parser() -> MuhParser:
    """Execute the main script."""
    parser = MuhParser(
        description="BIDS app using deepMReye to decode eye motion for fMRI time series data.",
        epilog="""
        For a more readable version of this help section,
        see the online https://bidsmreye.readthedocs.io/.
        """,
    )
    parser.add_argument(
        "bids_dir",
        help="""
        The directory with the input dataset formatted according to the BIDS standard.
        """,
    )
    parser.add_argument(
        "output_dir",
        help="""
        The directory where the output files will be stored.
        """,
    )
    parser.add_argument(
        "analysis_level",
        help="""Level of the analysis that will be performed.
        Multiple participant level analyses can be run independently (in parallel)
        using the same output_dir.
        """,
        choices=["participant"],
        default="participant",
    )
    parser.add_argument(
        "--action",
        help="""
        What action to perform.
        """,
        choices=["all", "prepare", "generalize"],
        default="all",
    )
    parser.add_argument(
        "--participant_label",
        help="""
        The label(s) of the participant(s) that should be analyzed.
        The label corresponds to sub-<participant_label> from the BIDS spec
        (so it does not include "sub-").

        If this parameter is not provided, all subjects will be analyzed.
        Multiple participants can be specified with a space separated list.
        """,
        nargs="+",
    )
    parser.add_argument(
        "--task",
        help="""
        The label of the task that will be analyzed.

        The label corresponds to task-<task_label> from the BIDS spec
        so it does not include "task-").
        """,
        nargs="+",
    )
    parser.add_argument(
        "--run",
        help="""
        The label of the run that will be analyzed.

        The label corresponds to run-<task_label> from the BIDS spec
        so it does not include "run-").
        """,
        nargs="+",
    )
    parser.add_argument(
        "--space",
        help="""
        The label of the space that will be analyzed.

        The label corresponds to space-<space_label> from the BIDS spec
        (so it does not include "space-").
        """,
        nargs="+",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="log_level",
        action="append_const",
        const=-1,
    )
    parser.add_argument("--debug", help="Switch to debug mode", action="store_true")
    parser.add_argument(
        "--reset_database",
        help="""
        Resets the database of the input dataset.
        """,
        action="store_true",
    )
    parser.add_argument(
        "--bids_filter_file",
        help="""
        A JSON file describing custom BIDS input filters using PyBIDS.
        For further details, please check out TBD.
        """,
    )
    parser.add_argument(
        "--version",
        action="version",
        help="show program's version number and exit",
        version=f"\nbidsMReye version {__version__}\n",
    )
    # TODO make it possible to pass path to a model ?
    prepare_only = parser.add_argument_group("prepare only arguments")
    prepare_only.add_argument(
        "--non_linear_coreg",
        help="""
        Uses a more aggressive (and non-linear) alignment procedure to the deepmreye template.
        """,
        action="store_true",
    )
    # TODO make it possible to pass path to a model ?
    generalize_only = parser.add_argument_group("generalize only arguments")
    generalize_only.add_argument(
        "--model",
        help="model to use",
        choices=available_models(),
        default=default_model(),
    )

    return parser


def args_to_dict(args: argparse.Namespace) -> dict[str, Any]:
    """Converts a argparse.Namespace object to a dictionary."""
    return vars(args)



from __future__ import annotations

import shutil
from pathlib import Path

from bidsmreye.generalize import convert_confounds
from bidsmreye.utils import get_dataset_layout


def test_convert_confounds():

    output_location = Path().resolve()
    output_location = output_location.joinpath("tests", "data", "bidsmreye")

    layout_out = get_dataset_layout(output_location)

    file = output_location.joinpath(
        "sub-01",
        "func",
        "sub-01_task-nback_space-MNI152NLin2009cAsym_desc-bidsmreye_confounds.npy",
    )
    shutil.copy(file, file.with_suffix(".bak"))

    confound_name = convert_confounds(layout_out, file)

    assert confound_name.is_file()

    confound_name.unlink()
    shutil.copy(file.with_suffix(".bak"), file)
    file.with_suffix(".bak").unlink()



"""Top-level package for bidsMReye."""
from __future__ import annotations

__author__ = """Remi Gau"""
__email__ = "remi.gau@gmail.com"

from . import _version

__version__ = _version.get_versions()["version"]



"""TODO."""
from __future__ import annotations

import logging
import os
import warnings
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
from bids import BIDSLayout  # type: ignore
from deepmreye import analyse
from deepmreye import train
from deepmreye.util import data_generator
from deepmreye.util import model_opts
from rich import print

from bidsmreye.quality_control import quality_control
from bidsmreye.utils import add_sidecar_in_root
from bidsmreye.utils import check_layout
from bidsmreye.utils import Config
from bidsmreye.utils import create_bidsname
from bidsmreye.utils import create_dir_for_file
from bidsmreye.utils import get_dataset_layout
from bidsmreye.utils import list_subjects
from bidsmreye.utils import move_file
from bidsmreye.utils import set_this_filter

log = logging.getLogger("bidsmreye")


def create_and_save_figure(
    layout_out: BIDSLayout, file: str, evaluation: Any, scores: Any
) -> None:
    """Generate a figure for the eye motion timeseries.

    Unused but keeping to help with plotting of datasets with with training data

    :param layout_out: Output dataset layout.
    :type  layout_out: BIDSLayout

    :param file:
    :type  file: str

    :param evaluation: see ``deepmreye.train.evaluate_model``
    :type  evaluation: _type_

    :param scores: see ``deepmreye.train.evaluate_model``
    :type  scores: _type_
    """
    fig = analyse.visualise_predictions_slider(
        evaluation,
        scores,
        color="rgb(0, 150, 175)",
        bg_color="rgb(255,255,255)",
        ylim=[-11, 11],
    )
    if log.isEnabledFor(logging.DEBUG):
        fig.show()

    confound_svg = create_bidsname(layout_out, file, "confounds_svg")
    create_dir_for_file(confound_svg)
    fig.write_image(confound_svg)


def convert_confounds(layout_out: BIDSLayout, file: str | Path) -> Path:
    """Convert numpy output to TSV.

    :param layout_out: pybids layout to of the dataset to act on.
    :type layout_out: BIDSLayout

    :param file: File to generate the confounds for.
    :type file: Union[str, Path]

    :return: Name of the file generated.
    :rtype: Path

    This function should preferably work on a single file
    but should still be able to unpack the results from a numpy file
    with results from multiple files.
    """
    confound_numpy = create_bidsname(layout_out, file, "confounds_numpy")

    content = np.load(
        file=confound_numpy,
        allow_pickle=True,
    )

    evaluation = content.item(0)
    for key, item in evaluation.items():
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            this_pred = np.nanmedian(item["pred_y"], axis=1)

        confound_name = create_bidsname(layout_out, Path(key + "p"), "confounds_tsv")

        log.info(f"Saving eye gaze data to {confound_name.relative_to(layout_out.root)}")

        pd.DataFrame(this_pred).to_csv(
            confound_name,
            sep="\t",
            header=["eye1_x_coordinate", "eye1_y_coordinate"],
            index=None,
        )

    log.debug(f"Removing {confound_numpy.relative_to(layout_out.root)}")
    os.remove(confound_numpy)

    return confound_name


def create_confounds_tsv(layout_out: BIDSLayout, file: str, subject_label: str) -> None:
    """Generate a TSV file for the eye motion timeseries.

    :param layout_out:
    :type layout_out: BIDSLayout

    :param file:
    :type file: str

    :param subject_label:
    :type subject_label: str
    """
    confound_numpy = create_bidsname(layout_out, file, "confounds_numpy")

    source_file = Path(layout_out.root).joinpath(
        f"sub-{subject_label}", "results_tmp.npy"
    )

    move_file(
        source_file,
        confound_numpy,
    )

    convert_confounds(layout_out, file)


def process_subject(cfg: Config, layout_out: BIDSLayout, subject_label: str) -> None:
    """Run generalize for one subject.

    :param cfg: Configuration object
    :type cfg: Config

    :param layout_out:
    :type layout_out: BIDSLayout

    :param subject_label:
    :type subject_label: str
    """
    log.info(f"Running subject: {subject_label}")

    this_filter = set_this_filter(cfg, subject_label, "no_label")

    data = layout_out.get(
        return_type="filename",
        regex_search=True,
        **this_filter,
    )

    to_print = [str(Path(x).relative_to(layout_out.root)) for x in data]
    log.debug(f"Found files\n{to_print}")

    for file in data:

        log.info(f"Processing file: {Path(file).name}")

        print("\n")
        generators = data_generator.create_generators([file], [file])
        generators = (*generators, [file], [file])
        print("\n")

        opts = model_opts.get_opts()

        (model, model_inference) = train.train_model(
            dataset="example_data",
            generators=generators,
            opts=opts,
            return_untrained=True,
        )
        model_inference.load_weights(cfg.model_weights_file)

        verbose = 0
        if log.isEnabledFor(logging.DEBUG):
            verbose = 2
        elif log.isEnabledFor(logging.INFO):
            verbose = 1

        (evaluation, scores) = train.evaluate_model(
            dataset="tmp",
            model=model_inference,
            generators=generators,
            save=True,
            model_path=f"{layout_out.root}/sub-{subject_label}/",
            model_description="",
            verbose=verbose,
            percentile_cut=80,
        )

        create_confounds_tsv(layout_out, file, subject_label)


def generalize(cfg: Config) -> None:
    """Apply model weights to new data.

    :param cfg: Configuration object
    :type cfg: Config
    """
    log.info("GENERALIZING")
    log.info(f"Using model: {cfg.model_weights_file}")

    layout_out = get_dataset_layout(cfg.output_folder)
    check_layout(cfg, layout_out)

    add_sidecar_in_root(layout_out)

    subjects = list_subjects(cfg, layout_out)

    for subject_label in subjects:

        process_subject(cfg, layout_out, subject_label)

    quality_control(cfg)



#!/usr/bin/env python
from __future__ import annotations

E402 = "E402: module level import not at top of file"

# see https://github.com/python-versioneer/python-versioneer/issues/249#issuecomment-1038184056
import sys

sys.path.insert(0, ".")


from setuptools import setup

import versioneer


SETUP_REQUIRES = ["setuptools >= 30.3.0"]
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []

if __name__ == "__main__":

    setup(
        setup_requires=SETUP_REQUIRES,
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
    )



from __future__ import annotations

from bidsmreye.bidsmreye import args_to_dict
from bidsmreye.bidsmreye import common_parser


def test_parser() -> None:
    parser = common_parser()
    args, unknowns = parser.parse_known_args(
        [
            "/path/to/bids",
            "/path/to/output",
            "participant",
            "--action",
            "prepare",
            "--task",
            "foo",
            "bar",
            "--non_linear_coreg",
        ]
    )

    assert args.task == ["foo", "bar"]
    assert args.non_linear_coreg


def test_parser_basic() -> None:

    parser = common_parser()

    assert (
        parser.description
        == "BIDS app using deepMReye to decode eye motion for fMRI time series data."
    )

    args, unknowns = parser.parse_known_args(
        [
            "/path/to/bids",
            "/path/to/output",
            "participant",
            "--action",
            "prepare",
            "--task",
            "foo",
            "bar",
        ]
    )

    assert args.task == ["foo", "bar"]
    assert args.non_linear_coreg == False



"""TODO."""
from __future__ import annotations

import json
import logging
import os
import re
import shutil
import warnings
from pathlib import Path
from typing import Any

from attrs import asdict
from attrs import converters
from attrs import define
from attrs import field
from bids import BIDSLayout  # type: ignore
from rich.logging import RichHandler
from rich.traceback import install

from . import _version

log = logging.getLogger("bidsmreye")

__version__ = _version.get_versions()["version"]


def default_log_level() -> str:
    """Return default log level."""
    return "WARNING"


def log_levels() -> list[str]:
    """Return a list of log levels."""
    return ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


def default_model() -> str:
    """Return default model."""
    return "1to6"


def copy_license(output_dir: Path) -> Path:
    """Copy CCO license to output directory.

    :param output_dir:
    :type output_dir: Path
    """
    input_file = str(Path(__file__).parent.joinpath("templates", "CCO"))
    output_file = output_dir.joinpath("LICENSE")
    create_dir_if_absent(output_dir)
    if not output_dir.joinpath("LICENSE").is_file():
        shutil.copy(input_file, output_dir)
        move_file(output_dir.joinpath("CCO"), output_file)
    return output_file


def add_sidecar_in_root(layout_out: BIDSLayout) -> None:
    content = {
        "StartTime": 0,
        "SampleCoordinateUnit": "degrees",
        "EnvironmentCoordinates": "center",
        "SampleCoordinateSystem": "gaze-on-screen",
        "RecordedEye": "both",
    }
    sidecar_name = Path(layout_out.root).joinpath("desc-bidsmreye_eyetrack.json")
    json.dump(content, open(sidecar_name, "w"), indent=4)


def create_sidecar(
    layout: BIDSLayout, filename: str, SamplingFrequency: float | None = None
) -> None:
    """Create sidecar for the eye motion timeseries."""
    if SamplingFrequency is None:
        SamplingFrequency = 0
    content = {
        "SamplingFrequency": SamplingFrequency,
    }
    sidecar_name = create_bidsname(layout, filename, "confounds_json")
    json.dump(content, open(sidecar_name, "w"), indent=4)


def available_models() -> list[str]:
    """Return a list of available models."""
    return [
        "1_guided_fixations",
        "2_pursuit",
        "3_openclosed",
        "3_pursuit",
        "4_pursuit",
        "5_free_viewing",
        "1to5",
        "1to6",
    ]


def set_this_filter(cfg: Config, subject_label: str, filter_type: str) -> dict[str, Any]:

    this_filter = cfg.bids_filter[filter_type]
    this_filter["suffix"] = return_regex(this_filter["suffix"])
    this_filter["task"] = return_regex(cfg.task)
    this_filter["space"] = return_regex(cfg.space)
    this_filter["subject"] = subject_label
    if cfg.run:
        this_filter["run"] = return_regex(cfg.run)

    log.debug(f"Looking for files with filter\n{this_filter}")

    return this_filter


def bidsmreye_log(name: str | None = None) -> logging.Logger:
    """Create log.

    :param name: _description_, defaults to None
    :type name: _type_, optional

    :return: _description_
    :rtype: _type_
    """
    # let rich print the traceback
    install(show_locals=True)

    FORMAT = "bidsMReye - %(asctime)s - %(message)s"

    log_level = default_log_level()

    if not name:
        name = "rich"

    logging.basicConfig(
        level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
    )

    return logging.getLogger(name)


@define
class Config:
    """Set up config and check that all required fields are set.

    :raises ValueError: _description_
    :raises RuntimeError: _description_

    :return: _description_
    :rtype: _type_
    """

    input_folder = field(default=None, converter=Path)

    @input_folder.validator
    def _check_input_folder(self, attribute: str, value: Path) -> None:
        if not value.is_dir:  # type: ignore
            raise ValueError(f"Input_folder must be an existing directory:\n{value}.")

    output_folder: Path = field(default=None, converter=Path)

    participant: Any | None = field(kw_only=True, default=None)

    space: Any | None = field(kw_only=True, default=None)
    task: Any | None = field(kw_only=True, default=None)
    run: Any | None = field(kw_only=True, default=None)

    model_weights_file: str | Path | None = field(kw_only=True, default=None)
    bids_filter: Any = field(kw_only=True, default=None)

    debug: str | bool | None = field(kw_only=True, default=None)
    reset_database: str | bool | None = field(kw_only=True, default=None)
    non_linear_coreg: bool = field(kw_only=True, default=False)

    has_GPU: bool = False

    def __attrs_post_init__(self) -> None:
        """Check that output_folder exists and gets info from layout if not specified."""
        os.environ["CUDA_VISIBLE_DEVICES"] = "0" if self.has_GPU else ""

        if not self.debug:
            self.debug = False
        if not isinstance(self.debug, (bool)):
            self.debug = converters.to_bool(self.debug)

        if not self.reset_database:
            self.reset_database = False
        if not isinstance(self.reset_database, (bool)):
            self.reset_database = converters.to_bool(self.reset_database)

        if not self.run:
            self.run = []

        # TODO test for passing bids_filter
        if not self.bids_filter:
            self.bids_filter = get_bids_filter_config()

        self.output_folder = self.output_folder.joinpath("bidsmreye")
        if not self.output_folder:
            self.output_folder.mkdir(parents=True, exist_ok=True)

        database_path = self.input_folder.joinpath("pybids_db")

        layout_in = BIDSLayout(
            self.input_folder,
            validate=False,
            derivatives=False,
            database_path=database_path,
            reset_database=self.reset_database,
        )

        if not database_path.is_dir():
            layout_in.save(database_path)

        self.check_argument(attribute="participant", layout_in=layout_in)
        self.check_argument(attribute="task", layout_in=layout_in)
        self.check_argument(attribute="run", layout_in=layout_in)
        self.check_argument(attribute="space", layout_in=layout_in)

    def check_argument(self, attribute: str, layout_in: BIDSLayout) -> Config:
        """Check an attribute value compared to a the input dataset content.

        :param attribute:
        :type attribute: _type_

        :param layout_in:
        :type layout_in: BIDSLayout

        :raises RuntimeError:

        :return:
        :rtype: _type_
        """
        if attribute == "participant":
            value = layout_in.get_subjects()
        elif attribute == "task":
            value = layout_in.get_tasks()
        elif attribute in {"space", "run"}:
            value = layout_in.get(return_type="id", target=attribute, datatype="func")

        self.listify(attribute)

        # convert all run values to integers
        if attribute in {"run"}:
            for i, j in enumerate(value):
                value[i] = int(j)
            tmp = [int(j) for j in getattr(self, attribute)]
            setattr(self, attribute, tmp)

        # keep only values that are intersection of requested values
        # and those present in the dataset
        if getattr(self, attribute):
            if missing_values := list(set(getattr(self, attribute)) - set(value)):
                warnings.warn(
                    f"{attribute}(s) {missing_values} not found in {self.input_folder}"
                )
            value = list(set(getattr(self, attribute)) & set(value))

        setattr(self, attribute, value)

        # run can be empty if run entity is not used
        if attribute not in ["run"] and not getattr(self, attribute):
            raise RuntimeError(f"No {attribute} not found in {self.input_folder}")

        return self

    def listify(self, attribute: str) -> Config:
        """Convert attribute to list if not already."""
        if getattr(self, attribute) and not isinstance(getattr(self, attribute), list):
            setattr(self, attribute, [getattr(self, attribute)])

        return self


def config_to_dict(cfg: Config) -> dict[str, Any]:
    """Convert a config to a dictionary.

    :param cfg:
    :type cfg: _type_

    :return:
    :rtype: _type_
    """
    dict_cfg = asdict(cfg)
    for key, value in dict_cfg.items():
        if isinstance(value, Path):
            dict_cfg[key] = str(value)
    return dict_cfg


def move_file(input: Path, output: Path) -> None:
    """Move or rename a file and create target directory if it does not exist.

    Should work even the source and target names are on different file systems.

    :param input:File to move.
    :type input: Path

    :param output:
    :type output: Path

    :param root: Optional. If specified, the printed path will be relative to this path.
    :type root: Path
    """
    log.debug(f"{input.resolve()} --> {output.resolve()}")
    create_dir_for_file(output)
    shutil.copy(input, output)
    input.unlink()


def create_dir_if_absent(output_path: str | Path) -> None:
    """Create a path if it does not exist.

    :param output_path:
    :type output_path: Union[str, Path]
    """
    if isinstance(output_path, str):
        output_path = Path(output_path)
    if not output_path.is_dir():
        log.debug(f"Creating dir: {output_path}")
    output_path.mkdir(parents=True, exist_ok=True)


def create_dir_for_file(file: Path) -> None:
    """Create the path to a file if it does not exist.

    :param file:
    :type file: Path
    """
    output_path = file.resolve().parent
    create_dir_if_absent(output_path)

    # TODO refactor with create_dir_if_absent


def return_regex(value: str | list[str] | None) -> str | None:
    """Return the regular expression for a string or a list of strings.

    :param value:
    :type value: Union[str, Optional[list]]

    :return:
    :rtype: Optional[str]
    """
    if isinstance(value, str):
        if value[0] != "^":
            value = f"^{value}"
        if not value.endswith("$"):
            value = f"{value}$"

    if isinstance(value, list):
        new_value = ""
        for i in value:
            new_value = f"{new_value}{return_regex(i)}|"
        value = new_value[:-1]

    return value


def list_subjects(cfg: Config, layout: BIDSLayout) -> list[str]:
    """List subject in a BIDS dataset for a given Config.

    :param cfg: Configuration object
    :type cfg: Config

    :param layout: BIDSLayout of the dataset.
    :type layout: BIDSLayout

    :raises RuntimeError: _description_

    :return: _description_
    :rtype: list
    """
    subjects = layout.get(return_type="id", target="subject", subject=cfg.participant)

    if subjects == [] or subjects is None:
        raise RuntimeError("No subject found")

    if cfg.debug:
        subjects = [subjects[0]]
        log.debug("Running first subject only.")

    log.info(f"processing subjects: {subjects}")

    return subjects


def get_deepmreye_filename(
    layout: BIDSLayout, img: str, filetype: str | None = None
) -> Path:
    """Get deepmreye filename.

    :param layout: BIDSLayout of the dataset.
    :type layout: BIDSLayout

    :param img: _description_
    :type img: str

    :param filetype: Any of the following: None, "mask", "report"D defaults to None
    :type filetype: str, optional

    :raises ValueError: _description_

    :return: _description_
    :rtype: Path
    """
    if not img:
        raise ValueError("No file")

    if isinstance(img, (list)):
        img = img[0]

    bf = layout.get_file(img)
    filename = bf.filename

    filename = return_deepmreye_output_filename(filename, filetype)

    filefolder = Path(img).parent
    filefolder = filefolder.joinpath(filename)

    return Path(filefolder).resolve()


def return_deepmreye_output_filename(filename: str, filetype: str | None = None) -> str:
    """Return deepmreye output filename.

    :param filename:
    :type filename: str

    :param filetype: Any of the following: None, "mask", "report". defaults to None
    :type filetype: str, optional

    :return: _description_
    :rtype: str
    """
    if filetype is None:
        pass
    elif filetype == "mask":
        filename = "mask_" + re.sub(r"\.nii.*", ".p", filename)
    elif filetype == "report":
        filename = "report_" + re.sub(r"\.nii.*", ".html", filename)

    return filename


def get_dataset_layout(
    dataset_path: str | Path,
    config: dict[str, str] | None = None,
    use_database: bool = False,
) -> BIDSLayout:
    """Return a BIDSLayout object for the dataset at the given path.

    :param dataset_path: Path to the dataset.
    :type dataset_path: Union[str, Path]

    :param config: Pybids config to use. Defaults to None.
    :type config: Optional[dict], optional

    :param use_database: Defaults to False
    :type use_database: bool, optional

    :return: _description_
    :rtype: BIDSLayout
    """
    if isinstance(dataset_path, str):
        dataset_path = Path(dataset_path)
    create_dir_if_absent(dataset_path)

    if config is None:
        pybids_config = get_pybids_config()

    log.info(f"indexing {dataset_path}")

    if not use_database:
        return BIDSLayout(
            dataset_path,
            validate=False,
            derivatives=False,
            config=pybids_config,
        )

    database_path = dataset_path.joinpath("pybids_db")
    return BIDSLayout(
        dataset_path,
        validate=False,
        derivatives=False,
        config=pybids_config,
        database_path=database_path,
    )


def write_dataset_description(layout: BIDSLayout) -> None:
    """Add a dataset_description.json to a BIDS dataset.

    :param layout: BIDSLayout of the dataset to update.
    :type layout: BIDSLayout
    """
    output_file = Path(layout.root).joinpath("dataset_description.json")

    with open(output_file, "w") as ff:
        json.dump(layout.dataset_description, ff, indent=4)


def set_dataset_description(layout: BIDSLayout, is_derivative: bool = True) -> BIDSLayout:
    """Add dataset description to a layout.

    :param layout: _description_
    :type layout: BIDSLayout

    :param is_derivative: Defaults to True
    :type is_derivative: bool, optional

    :return: Updated BIDSLayout of the dataset
    :rtype: BIDSLayout
    """
    data: dict[str, Any] = {
        "Name": "dataset name",
        "BIDSVersion": "1.7.0",
        "DatasetType": "raw",
        "License": "CCO",
        "Authors": ["", ""],
        "Acknowledgements": "Special thanks to ",
        "HowToAcknowledge": """Please cite this paper: Frey, M., Nau, M. & Doeller, C.F.
Magnetic resonance-based eye tracking using deep neural networks.
Nat Neurosci 24, 1772-1779 (2021).
https://doi.org/10.1038/s41593-021-00947-w""",
        "Funding": ["", ""],
        "ReferencesAndLinks": ["https://doi.org/10.1038/s41593-021-00947-w"],
        "DatasetDOI": "doi:",
    }

    if is_derivative:
        data["GeneratedBy"] = [
            {
                "Name": "bidsMReye",
                "Version": __version__,
                "Container": {"Type": "", "Tag": ""},
                "Description": "",
                "CodeURL": "https://github.com/cpp-lln-lab/bidsMReye.git",
            },
        ]

        data["SourceDatasets"] = [
            {
                "DOI": "doi:",
                "URL": "",
                "Version": "",
            }
        ]

    layout.dataset_description = data

    return layout


def init_derivatives_layout(output_location: Path) -> BIDSLayout:
    """Initialize a derivatives dataset and returns its layout.

    :param output_location:
    :type output_location: Path

    :return:
    :rtype: BIDSLayout
    """
    layout_out = get_dataset_layout(output_location)
    layout_out = set_dataset_description(layout_out)
    layout_out.dataset_description["DatasetType"] = "derivative"
    layout_out.dataset_description["GeneratedBy"][0]["Name"] = "bidsmreye"
    write_dataset_description(layout_out)
    return layout_out


def get_bidsname_config(config_file: Path | None = None) -> dict[str, str]:
    """Load configuration for naming output BIDS files.

    :param config_file: Defaults to None
    :type config_file: Path, optional

    :return: Config as a dictionary.
    :rtype: dict

    See the Path construction demo in the pybids tutorial.

    https://github.com/bids-standard/pybids/blob/master/examples/pybids_tutorial.ipynb
    """
    default = "config_bidsname.json"
    return get_config(config_file, default)


def get_pybids_config(config_file: Path | None = None) -> dict[str, str]:
    """Load pybids configuration.

    :param config_file: Defaults to None
    :type config_file: Path, optional
    :return: _description_
    :rtype: dict

    Pybids configs are stored in the layout module.

    https://github.com/bids-standard/pybids/tree/master/bids/layout/config
    """
    default = "config_pybids.json"
    return get_config(config_file, default)


def get_bids_filter_config(config_file: Path | None = None) -> dict[str, Any]:
    """Load the bids filter file config.

    :param config_file: Config to load. Defaults to None.
    :type config_file: Path, optional

    :return: _description_
    :rtype: dict
    """
    default = "default_filter_file.json"
    return get_config(config_file, default)


def get_config(config_file: Path | None = None, default: str = "") -> dict[str, str]:
    """Load a config stored in a JSON.

    :param config_file: File to load. Defaults to None.
                        Will look into the config directory if None.
    :type config_file: Path, optional

    :param default: Default file to load. Defaults to ""
    :type default: str, optional

    :raises FileNotFoundError: _description_

    :return: Config as a dictionary.
    :rtype: dict
    """
    if config_file is None or not Path(config_file).exists():
        my_path = Path(__file__).resolve().parent.joinpath("config")
        config_file = my_path.joinpath(default)

    if config_file is None or not Path(config_file).exists():
        raise FileNotFoundError(f"Config file {config_file} not found")

    with open(config_file) as ff:
        return json.load(ff)


def create_bidsname(
    layout: BIDSLayout, filename: dict[str, str] | str | Path, filetype: str
) -> Path:
    """Return a BIDS valid filename for layout and a filename or a dict of BIDS entities.

    :param layout: BIDSLayout of the dataset.
    :type layout: BIDSLayout

    :param filename: Dictionary of BIDS entities or a Path to a file.
    :type filename: Union[dict, str, Path]

    :param filetype: One of the file type available in the BIDS name config.
    :type filetype: str

    :raises TypeError:

    :return:
    :rtype: Path
    """
    if isinstance(filename, dict):
        entities = filename
    elif isinstance(filename, (Path, str)):
        entities = layout.parse_file_entities(filename)
    else:
        raise TypeError(f"filename must be a dict or a Path, not {type(filename)}")

    bids_name_config = get_bidsname_config()

    output_file = layout.build_path(entities, bids_name_config[filetype], validate=False)

    output_file = Path(layout.root).joinpath(output_file)

    return output_file.resolve()


def check_layout(cfg: Config, layout: BIDSLayout) -> None:
    """Check layout.

    :param cfg: Configuration object
    :type cfg: Config

    :param layout: BIDSLayout of the dataset.
    :type layout: BIDSLayout

    :raises RuntimeError: _description_
    :raises RuntimeError: _description_
    """
    desc = layout.get_dataset_description()
    if (
        "DatasetType" not in desc
        and "PipelineDescription" not in desc
        or "DatasetType" in desc
        and desc["DatasetType"] != "derivative"
    ):
        raise RuntimeError("Input dataset should be BIDS derivative")

    this_filter = get_bids_filter_config()["bold"]

    if "GeneratedBy" in desc:
        generated_by = desc["GeneratedBy"]
    elif "PipelineDescription" in desc:
        generated_by = desc["PipelineDescription"]

    if isinstance(generated_by, list):
        generated_by = generated_by[0]

    if generated_by["Name"].lower() == "bidsmreye":
        this_filter = get_bids_filter_config()["mask"]

    this_filter["task"] = cfg.task
    this_filter["space"] = cfg.space
    this_filter["run"] = cfg.run

    log.debug(f"Looking for files with filter\n{this_filter}")

    bf = layout.get(
        return_type="filename",
        regex_search=True,
        **this_filter,
    )

    if bf == []:
        raise RuntimeError(
            f"""
            Input dataset {layout.root} does not have any data to process for filter\n{this_filter}
            """
        )



"""Write method section."""
from __future__ import annotations

import shutil
import warnings
from pathlib import Path

import chevron

from . import _version
from bidsmreye.utils import available_models
from bidsmreye.utils import create_dir_for_file
from bidsmreye.utils import default_model


__version__ = _version.get_versions()["version"]


def methods(output_dir: str | Path = Path("."), model_name: str | None = None) -> Path:
    """Write method section.

    :param output_dir: Defaults to Path(".")
    :type  output_dir: Union[str, Path], optional

    :param model_name: Defaults to None.
    :type  model_name: str, optional

    :return: Output file name.
    :rtype: Path
    """
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    output_dir = output_dir.joinpath("logs")

    output_file = output_dir.joinpath("CITATION.md")
    create_dir_for_file(output_file)

    bib_file = str(Path(__file__).parent.joinpath("templates", "CITATION.bib"))
    shutil.copy(bib_file, output_dir)

    if not model_name:
        model_name = default_model()

    is_known_models = False
    is_default_model = False
    if model_name in available_models():
        is_known_models = True
        if model_name == default_model():
            is_default_model = True

    if not is_known_models:
        warnings.warn(f"{model_name} is not a known model name.")

    template_file = str(Path(__file__).parent.joinpath("templates", "CITATION.mustache"))
    with open(template_file) as template:

        output = chevron.render(
            template=template,
            data={
                "version": __version__,
                "model_name": model_name,
                "is_default_model": is_default_model,
                "is_known_models": is_known_models,
            },
            warn=True,
        )

    output_file.write_text(output)

    return output_file
